require "matrix"
require "uranai/data"
require "uranai/fitting_parameter"

module Uranai
  class LinearRegression
    def initialize(csv)
      @data = Uranai::Data.new(csv)
    end

    def compute_cost(theta: default_theta)
      matrix_theta = matrix_theta(theta)
      ((matrix_x * matrix_theta) - matrix_y)
        .map { |element| element ** 2 }
        .inject(0.0) { |sum,i| sum + i }
        ./ (2 * data.training_example_size)
    end

    def fitting_params_with_gradient_descent(options = {})
      options.keep_if { |k,_| [:alpha, :num_iters, :theta, :normalize].include?(k) }
      FittingParameter.new(gradient_descent(options))
    end
    alias :fitting_params :fitting_params_with_gradient_descent

    def fitting_params_with_normal_equations
      FittingParameter.new(normal_equations)
    end

    private

    attr_reader :data

    def default_theta
      [[0] * data.column_size]
    end

    def normalize_theta(theta)
      [theta.flatten]
    end

    def matrix_x
      data.matrix_x
    end

    def normalized_matrix_x
      data.normalized_matrix_x
    end

    def matrix_y
      data.matrix_y
    end

    def matrix_theta(theta)
      Matrix.columns(normalize_theta(theta))
    end

    def training_example_size
      data.training_example_size
    end

    # TODO: Add an option for normalizing feature data (x data)
    def gradient_descent(alpha: 0.01, num_iters: 1500, theta: default_theta, normalize: false)
      matrix_theta = matrix_theta(theta)
      x = normalize ? normalized_matrix_x : matrix_x
      num_iters.times do
        matrix_theta -=
          (x.transpose * ((x * matrix_theta) - matrix_y)) * (alpha / training_example_size)
      end
      matrix_theta.to_a.flatten
    end

    def normal_equations
      ((matrix_x.transpose * matrix_x).inverse * matrix_x.transpose * matrix_y)
        .to_a.flatten
    end
  end
end
